#!/usr/bin/perl -w

use strict;
use Parse::AccessLogEntry;
use Date::Parse;
use Date::Format;
use Time::Piece;
use Time::Seconds;
use Number::Format qw(:subs);
use DBI;

#  $Hashref->{host}    client ip of the request
#  $Hashref->{user}    user logged in ("-" for none)
#  $Hashref->{date}    date of the request
#  $Hashref->{time}    server time of the request
#  $Hashref->{diffgmt} server offset from GMT
#  $Hashref->{rtype}   type of request (GET, POST, etc)
#  $Hashref->{file}    file requested
#  $Hashref->{proto}   protocol used (HTTP/1.1, etc)
#  $Hashref->{code}    code returned by apache (200, 304, etc)
#  $Hashref->{bytes}   number of bytes returned to the client
#  $Hashref->{refer}   referrer
#  $Hashref->{agent}   user-agent

###############################
### configuration variables ###
###############################
my $combined_log = "/var/log/httpd-tracker.log"; # apache combined log format
my $sample_size = @ARGV >= 1 ? int($ARGV[0]) : 75000;   # used with 'tail'
my $dbname = "dbname";                     # name of MySQL database
my $dbuser = "voracity";                        # user name for DB
my $dbpass = "dbpasswd";                        # password for user

my $p = Parse::AccessLogEntry::new();

my $first_ts;
my $last_ts;
my $tz;

my %hits = ();
my %bw = ();

my %cats = ( 
            "tracker" =>    qr!^/announce\.php$! ,
            "scrape" =>     qr!^/(announce\.php/)?scrape(\.php)?$! ,
#            "download" =>   qr!^/download\.php! ,
#            "image" =>      qr!\.(gif|jpe?g|png)$!i , 
#            "index" =>      qr!^/(index\.php)?$! ,
#            "css" =>        qr!\.css$! ,
           );

foreach ("total", "other", keys %cats) {
    $hits{$_} = 0;
    $bw{$_} = 0;
}


open(FH, "tail -$sample_size $combined_log |") or die("can't open tail -$sample_size $combined_log: $!");
                 
while( <FH> ) {
    print STDERR "$.\r" if $. % 1000 == 0;
    my $l = $p->parse( $_ );

#    my $ts = str2time($l->{date} . " " . $l->{time} . " " . $l->{diffgmt});
    my $ts = Time::Piece->strptime($l->{date} . " " . $l->{time}, "%d/%b/%Y %H:%M:%S");
    
    $tz = $l->{diffgmt} unless defined $tz;
    $first_ts = $ts unless defined $first_ts;
    $last_ts = $ts;
    
    my ($uri, $params) = split(/\?/, $l->{file}, 2);

    # default to "other" if no RE matches
    my $c = "other";

    # determine which category this uri falls under
    foreach (keys %cats) {
        if( $uri =~ /$cats{$_}/ ) {
            $c = $_;
            last;
        }
    }

    # account for the hit in that category
    $hits{$c}++;
    $hits{"total"}++;

    my $bytes = ($l->{bytes} =~ /^\d+$/ ? $l->{bytes} : 0);
    $bw{$c} += $bytes;
    $bw{"total"} += $bytes;
}

close(FH);

#printf "Begin statistic period: %s\n", time2str("%a %b %e %T %Y", $first_ts);
#printf "End   statistic period: %s\n\n", time2str("%a %b %e %T %Y", $last_ts);

printf "Begin sampling period: %s $tz\n", $first_ts->cdate;
printf "End   sampling period: %s $tz\n\n", $last_ts->cdate;

my $period = $last_ts - $first_ts;

my @labels = ("Type", "Hits", "Bytes", "Hits/s", "Hits/day", "Bytes/hit", "Bytes/day", "(%)");

my $head = join(' ', map(sprintf("%-10.10s", $_), @labels));
my $div = "-" x length($head);
my $fmt = "%-10.10s " x scalar(@labels) . "\n";

print "$head\n$div\n";
foreach ((sort { $bw{$b} <=> $bw{$a} } (keys %cats, "other")), "total") {
    print "$div\n" if $_ eq "total";
    printf $fmt, 
                            $_, 
                            format_number($hits{$_}), 
                            format_bytes($bw{$_}, 3),
                            format_number($hits{$_} / $period, 2), 
                            format_number($hits{$_} / $period * 60*60*24, 0), 
                            format_bytes($bw{$_} / $hits{$_}, 3), 
                            format_bytes($bw{$_} / $period * 60*60*24, 3),
                            $_ ne "total" ? "(" . sprintf("%.1f", $bw{$_} / $bw{"total"} * 100) . "%)" : "",
                            ;
}


# now get some stats from the database

my $dbh = DBI->connect("dbi:mysql:$dbname", $dbuser, $dbpass, {RaiseError => 1});

my ($tot_peers, $uniq_peers) = $dbh->selectrow_array( qq{
    SELECT count(*) AS total_peers, count(DISTINCT ip) AS distinct_peers FROM peers
} );

my ($tot_torrents) = $dbh->selectrow_array( qq{
    SELECT count(*) AS total_torrents FROM torrents
} );

my ($visible_torrents) = $dbh->selectrow_array( qq{
    SELECT count(*) AS visible_torrents FROM torrents WHERE visible = 'yes' and banned = 'no'
} );

my ($tot_users) = $dbh->selectrow_array( qq{
    SELECT count(*) AS total_users FROM users
} );

my ($confirmed_users) = $dbh->selectrow_array( qq{
    SELECT count(*) AS total_users FROM users WHERE status = 'confirmed'
} );

$dbh->disconnect();

printf "\n\n%s peers connected, %s unique.\n%s torrents in database, %s visible.\n%s user accounts, %s confirmed.\n",
            map format_number($_), ($tot_peers, $uniq_peers, $tot_torrents, $visible_torrents, $tot_users, $confirmed_users);

# my $up = Time::Seconds->new(uptime());

# print STDERR "\ndebug: \$up = $up, uptime = " . uptime() . "\n";

# sample '10:49AM  up 29 days, 16:38, 2 users, load averages: 1.30, 1.92, 2.25'
my $up = `uptime`;
$up =~ m!up\s+(\d+)\s+days?,\s+(\d+):(\d+)!i;
my ($days, $hours, $mins) = ($1, $2, $3); 

#my $weeks = floor($up->weeks()); $up -= (ONE_WEEK * $weeks);
#my $days = floor($up->days()); $up -= (ONE_DAY * $days);
#my $hours = floor($up->hours()); $up -= (ONE_HOUR * $hours);

printf "The server was last rebooted %d day%s, %d hour%s, %d minute%s ago.\n", 
            $days, ($days == 1 ? "":"s"),
            $hours, ($hours == 1 ? "":"s"),
            $mins, ($mins == 1 ? "":"s"),
            ;
                        