#!/usr/bin/perl -w

use strict;
use DBI;
use Data::Dumper;
use GD::Graph::points;
use GD::Graph::lines;
use GD::Graph::area;
use GD::Graph::Data;
use GD::Graph::colour qw(:colours :lists :files :convert); 
use Date::Format;
use POSIX qw(floor ceil);
use HTML::Entities;
use Date::Format;
use Proc::Daemon;
use Time::HiRes qw(gettimeofday tv_interval);


###############################
### configuration variables ###
###############################
my $dbname = "dbname";                     # name of MySQL database
my $dbuser = "voracity";                        # user name for DB
my $dbpass = "dbpasswd";                        # password for user

my $outputdir = "/usr/local/www/data/stats";    # no trailing slash
my $fontname = "/home/admin/fonts/verdana.ttf"; # TTF font used in graph  
my $logfile = "/var/log/torstatsd.log";         # messages
my $pidfile = "/var/run/torstatsd.pid";         # process ID

my %colors = ( foreground => "#E5E5E5",         # axis and grid
               background => "#000000",         # main background (will be marked transparent)
               textlabels => "#ACA5E5",         # textual labels, i.e. series names
               axislabels => "#E5A5B2",         # values printed along axes, i.e. numbers
               legend =>     "#E5E5E5",         # text in the legend
               text =>       "#E5E5E5",         # all other text
               seeders =>    "#ACA5E5",         # the rest of these are the plot data colors:
               leechers =>   "#A5C4E5",
               total =>      "#E5E5E5",
               completes =>  "#ACA5E5",
               views =>      "#A5C4E5",
               hits =>       "#E5A5B2",
               min =>        "#ACA5E5",
               avg =>        "#E5E5E5",
               max =>        "#E5A5B2", 
             );               

map { add_colour( "$_" => [ hex2rgb( $colors{$_} ) ] ); } keys %colors;

my $bintime = 60*5;                             # number of seconds per sampling bin
my $update_interval = 60*60*4;                  # how often to regenerate graphs (seconds)
my $dither_secs = 60;                           # vary interval by +- this amount
my $num_y_ticks = 6;                            # number of axis tick subdivisions
my $num_x_ticks = 6;
my ($graph_width, $graph_height) = (640, 300);  # size of the graphs

# set options common to all graphs here:
my %graph_options = (
#    x_tick_number => 'auto',
    x_tick_number => $num_x_ticks,
    x_number_format => sub { time2str("%L/%d %H:%M", $_[0], "UTC"); },
    x_labels_vertical => 1,
    y_tick_number => $num_y_ticks,
    y_long_ticks => 1,
    y_min_value => 0,
    line_width => 3,
    t_margin => 10,
    b_margin => 10,
    r_margin => 10,
    axis_space => 15,
    text_space => 15,
    transparent => 1,
    interlaced => 0,
    bgclr => "background",
    fgclr => "foreground",
    boxclr => undef,
    labelclr => "textlabels",
    axislabelclr => "axislabels",
    legendclr => "legend",
    textclr => "text",
    legend_placement => 'BC',
#    markers => [7, 7],
#    marker_size => 1, 
);


########################
### global variables ###
########################
my $nrows;
my $secs_until;
my $done_flag = 0;

# make these global in an attempt to stem the bulging RSS as this runs over time
my $sl_data = GD::Graph::Data->new();       # seeders and leechers plot
my $cvh_data = GD::Graph::Data->new();      # completes, views, hits plot
my $mam_data = GD::Graph::Data->new();      # min-avg-max leecher completion plot


########################
### detach and begin ###
########################

Proc::Daemon::Init;
umask(022);


# write pid file, open log file
open (PID, ">", $pidfile) or die "Can\'t create pidfile \'$pidfile\': $!";
print PID "$$\n";
close PID;

open (STDERR, ">>", $logfile) or die "Can\'t open logfile \'$logfile\': $!";
autoflush STDERR 1;
do_log("torstatsd starting");


# initialize the database
my $dbh;
unless( $dbh = DBI->connect("dbi:mysql:$dbname", $dbuser, $dbpass, { RaiseError => 0, PrintError => 1 }) ) {
    do_log("can't connect to DB: $DBI::errstr");
    exit 1;
}

# prepare the statements
my ($sth_check_new, $sth_sched_new, $sth_mk_index, $sth_get_next, $sth_reschedule_getns, $sth_mk_graphs_getdata);
unless( 
    $sth_check_new = $dbh->prepare( q{
        SELECT id
        FROM torrents
        WHERE next_stats = '0000-00-00 00:00:00'
    }) and

    $sth_sched_new = $dbh->prepare( q{
        UPDATE torrents
        SET next_stats = NOW() + INTERVAL ? SECOND
        WHERE id = ?
    }) and

    $sth_mk_index = $dbh->prepare( q{
        SELECT id, name
        FROM torrents
    }) and

    $sth_get_next = $dbh->prepare( q{
        SELECT next_stats, (UNIX_TIMESTAMP(next_stats) - UNIX_TIMESTAMP( NOW() )) AS secs_until, id, name, UNIX_TIMESTAMP(last_stats_datapoint) AS ts
        FROM torrents
        WHERE visible = 'yes'
        ORDER BY secs_until ASC
        LIMIT 1
    }) and

    $sth_reschedule_getns = $dbh->prepare( q{
        SELECT next_stats
        FROM torrents
        WHERE id = ?
    }) and

    $sth_mk_graphs_getdata = $dbh->prepare( q{
        SELECT UNIX_TIMESTAMP(timestamp) AS ts, seeders, leechers, times_completed, views, hits, min_pct, avg_pct, max_pct
        FROM torrents_bookkeeping 
        WHERE torrent_id = ?
        ORDER BY timestamp DESC
    })

    )
{
    do_log("error preparing statement: $DBI::errstr");
    $done_flag = 1;
}



$SIG{INT} = sub { $done_flag = 1; };
$SIG{TERM} = sub { $done_flag = 1; };


#################
### main loop ###
#################

until( $done_flag ) {
    
    my ($next_stats, $id, $name, $lsd);
    
    ($next_stats, $secs_until, $id, $name, $lsd) = get_next();
    
    sleep $secs_until if($secs_until > 1);
#    do_log("main: warning: " . -($secs_until) . " seconds behind schedule") unless ($secs_until > -30);
    
    my $this_lsd = mk_graphs($id, $name, $lsd) || 1;        # in case there were 0 rows or something, use a default value of 1

    reschedule($id, $this_lsd);
}

################
### shutdown ###
################

do_log("torstatsd stopping");
close STDERR;
# unlink($pidfile);
exit 0;



#########################
### support functions ###
#########################

    
sub do_log {
    # spews a message with timestamp to the log file
    my @now = localtime(time);

    my $ts = strftime("[%d/%b/%Y:%X %z]", @now);
    my $msg = join(" ", @_);
    
    print STDERR "$ts $msg\n";
}


sub get_next {
    # returns (next_stats, number of seconds to wait until next event, ID of that torrent, name of torrent, last_stats_datapoint)

    unless( $sth_get_next->execute ) {
        do_log("get_next: error executing sth_get_next: $sth_get_next->errstr");
        return( "", 30, -1, "", 0 );
    }
    
    my @next_torrent;
    unless( @next_torrent = $sth_get_next->fetchrow_array ) {
        do_log("get_next: error fetching sth_get_next: $sth_get_next->errstr");
        return( "", 30, -1, "", 0 );
    }
        
    # has a new torrent been added that needs scheduling?
    if( $next_torrent[0] eq "0000-00-00 00:00:00" ) {
        check_new();
        
        # would recursion be an issue here?
        # what about a race condition?  It would occur if a torrent kept being added between check_new() and here
        # That would really be bad, so this should really enforce one level of recursion or something.
        return get_next();
    }        
    return @next_torrent;
}

sub check_new {
    # checks for torrents with 'next_stats' = 0000-00-00 00:00:00
    # and schedules them for stats-generation
    
    unless( $sth_check_new->execute ) {
        do_log("check_new: error executing sth_check_new: $sth_check_new->errstr");
        return;
    }
    
    my $ids;
    unless( $ids = $sth_check_new->fetchall_hashref("id") ) {
        do_log("check_new: error fetching: $sth_check_new->errstr");
        return;
    }

    return if( scalar(keys %$ids) < 1);
    
    # schedule them out evenly over the next update_interval
    my $stepsize = ceil($update_interval / scalar(keys %$ids));
    my $offset = $stepsize;
    
    foreach(keys %$ids) {
        unless( $sth_sched_new->execute( $offset, $_ ) == 1) {      
            do_log("check_new: warning: problem updating \'next_stats\' for torrent $_ ($sth_sched_new->errstr)");
        }
        $offset += $stepsize;
    }

    do_log("check_new: scheduled " . scalar( keys %$ids ) . " new torrents for stats.");
    mk_index();
}       

sub mk_index {
    # this creates index.html in the stats output dir

    unless( $sth_mk_index->execute ) {
        do_log("mk_index: error executing sth_mk_index: $sth_mk_index->errstr");
        return;
    }

    my $ids;
    unless( $ids = $sth_mk_index->fetchall_hashref("id") ) {
        do_log("mk_index: error fetching: $sth_mk_index->errstr");
        return;
    }
   
    unless( open(FH, ">", $outputdir . "/index.html") ) {
        do_log("mk_index: can't create $outputdir/index.html: $!");
        return;
    }
    
    print FH "<html><body text=\"#E5E5E5\" bgcolor=\"#000000\">\n";
    
    foreach (sort {$a <=> $b} keys %{ $ids }) {
        
        my $name = encode_entities($ids->{$_}->{name});
    
        print FH "<h1>$name ($_)</h1>\n" . 
                 "<h2>Seeders and Total Peers</h2>\n<img src=\"$_-sl.png\" width=\"$graph_width\" height=\"$graph_height\"><br>\n" .
                 "<h2>Completes, Views, Hits</h2>\n<img src=\"$_-cvh.png\" width=\"$graph_width\" height=\"$graph_height\"><br>\n" .
                 "<h2>Leecher Completion (Min, Average, Max)</h2>\n<img src=\"$_-mam.png\" width=\"$graph_width\" height=\"$graph_height\"><br>\n\n<hr>\n";
    
    }   
    print FH "</body></html>\n";
    close FH;
    do_log("mk_index: regenerated $outputdir/index.html");
}    
    


sub reschedule {
    # updates the 'next_stats' and 'last_stats_datapoint' fields
    my ($id, $lsd) = @_;
    
    unless( defined($id) && $id > 0 ) {
        do_log("reschedule: bad \$id parameter");
        return;
    }
    
    unless( defined($lsd) && $lsd > 0 ) {
        do_log("reschedule: bad \$last_stats_datapoint parameter: \'$lsd\'");
        return;
    }

    my $interval = $update_interval + floor(rand(2 * $dither_secs)) - $dither_secs;

    # this catches the case where a new torrent is added and has not been scheduled before mk_graphs() is called
    unless( $sth_reschedule_getns->execute( $id ) ) {
        do_log("reschedule: error executing sth_reschedule_getns: $sth_reschedule_getns->errstr");
        return;
    }
    
    my $ns;
    unless( $ns = $sth_reschedule_getns->fetchrow_array ) {
        do_log("reschedule: error fetching sth_reschedule_getns: $sth_reschedule_getns->errstr");
        return;
    }
    
    my $begintime = ( $ns eq "0000-00-00 00:00:00" ? "NOW()" : "next_stats" );

    # I don't think I can use placeholders here....
    my $rv = $dbh->do( qq{
        UPDATE torrents
        SET next_stats = $begintime + INTERVAL $interval SECOND, last_stats_datapoint = FROM_UNIXTIME($lsd)
        WHERE id = $id
    } );
    do_log("reschedule: warning: UPDATE should have returned 1, but got $rv") unless ( $rv == 1 );
}

    
sub mk_graphs {
    # passed the ID of a torrent, this queries the database for its stats and
    # creates the three plots
    # returns the highest datapoint timestamp for the torrent (this_lsd)
    # which is used by the calling process and gets set in the torrents DB by reschedule()
    my $time_start = [gettimeofday];
    my $this_lsd;                       # will store the "last stats datapoint" value based on the records in torrents_bookkeeping

    my $torrent_id = int shift;         # make sure it's numeric!
    my $torrent_name = shift;
    my $lsd = shift;                    # value of the 'last_stats_datapoint' field from torrents table

    # get data from the DB for our torrent
    unless( $sth_mk_graphs_getdata->execute( $torrent_id ) ) {
        do_log("mk_graphs: error executing sth_mk_graphs_getdata: $sth_mk_graphs_getdata->errstr");
        return $this_lsd;
    }
       
    # bind the result columns to these variable names
    my ($ts, $s, $l, $comp, $v, $h, $min, $avg, $max);    
    $sth_mk_graphs_getdata->bind_columns( \($ts, $s, $l, $comp, $v, $h, $min, $avg, $max) );

    # these will be used to "bin" the samples (based on the $bintime setting)
    my (%s_bins, %l_bins, %comp_bins, %v_bins, %h_bins, %min_bins, %avg_bins, %max_bins);

    my $recno;
    
    # get results, and store in the hash-bins   
    for( $recno = 0; $sth_mk_graphs_getdata->fetch; $recno++ ) {

        if( $recno == 0 ) {
        
            # the order of our results is DESC, so the first $ts will be the highest (latest) timestamp
            # this should be saved in the global var $this_lsd
            $this_lsd = $ts;

            # if nothing has changed since the last time we made this graph, then bail
            # $lsd is passed to us from main, and is retrieved via get_next()
            if( $this_lsd == $lsd ) {
                my $elap_secs = tv_interval $time_start, [gettimeofday];
        
                do_log(sprintf("mk_graphs: [%.3f s] graph is up to date for %s (%s)", $elap_secs, $torrent_id, $torrent_name));
        
                return $this_lsd;
            }
        }        

        my $binned = floor($ts/$bintime)*$bintime;
        push @{ $s_bins{$binned} }, $s;
        push @{ $l_bins{$binned} }, $l;

        push @{ $comp_bins{$binned} }, $comp;
        push @{ $v_bins{$binned} }, $v;
        push @{ $h_bins{$binned} }, $h;

#        if($min == -1 || $avg == -1 || $max == -1) {
        if($min == -1 || $avg == -1 || $max == -1 || ($min == 0 && $avg == 0 && $max == 0)) {
            # this means "no peers"
            # $gaps{$binned} = 1;
        } else {                    
            push @{ $min_bins{$binned} }, (1 - $max) * 100;     # in the DB, these are "percent to go", we graph "percent done"
            push @{ $avg_bins{$binned} }, (1 - $avg) * 100;
            push @{ $max_bins{$binned} }, (1 - $min) * 100;
        }
    }

    my $time_fetch = [gettimeofday];

    # check for error condition (recall that error and "no more rows" both return undef from fetch)
    if( $sth_mk_graphs_getdata->err ) {
        do_log("mk_graphs: torrent $torrent_id ($torrent_name) error fetching data: $sth_mk_graphs_getdata->errstr");
        return $this_lsd;
    }            

    # check for zero-rows result
    if( $recno == 0 ) {
        do_log("mk_graphs: torrent $torrent_id ($torrent_name) returned zero rows from torrents_bookkeeping");
        return $this_lsd;
    }

    # determine our y-axis scaling while we do this
#    my ($sl_max, $cvh_max) = (0, 0);

    # reset data objects for the 3 plots
    $sl_data->reset();
    $cvh_data->reset();
    $mam_data->reset();
    GD::Graph::Data->reset();
    
    # sort the timestamp bins and iterate
    my @bins = sort keys %s_bins;
    foreach my $bin (@bins) {
        
        my @averages = ( avg( $s_bins{$bin} ),
                         avg( $l_bins{$bin} ),
                         avg( $comp_bins{$bin} ),
                         avg( $v_bins{$bin} ),
                         avg( $h_bins{$bin} ),
                         exists($min_bins{$bin}) ? avg( $min_bins{$bin} ) : undef,
                         exists($avg_bins{$bin}) ? avg( $avg_bins{$bin} ) : undef,
                         exists($max_bins{$bin}) ? avg( $max_bins{$bin} ) : undef,
                        );
        
#        $sl_max = max([ $sl_max, @averages[0,1] ]);
#        $sl_max = max([ $sl_max, $averages[0] + $averages[1] ]);
#        $cvh_max = max([ $cvh_max, @averages[2..4] ]);
               
        $sl_data->add_point($bin, $averages[0], $averages[0] + $averages[1]);
#        $sl_data->add_point($bin, @averages[0,1]);
        $cvh_data->add_point($bin, @averages[2..4]);
        $mam_data->add_point($bin, @averages[5..7]);        
    }
    
    my $sl_max = max([ $sl_data->get_min_max_y_all(), 0.01 ]);
    my $cvh_max = max([ $cvh_data->get_min_max_y_all(), 0.01 ]);

#    do_log("mk_graphs: debug: \$sl_max = $sl_max, \$cvh_max = $cvh_max");

    my $basename = $outputdir . "/$torrent_id";

#    open(DBG, ">", "$basename-debug.txt") or do_log("mk_graphs: $!");
#    print DBG "\$sl_data\n" . Dumper($sl_data) . "\n\n\n\$cvh_data\n" . Dumper($cvh_data) . "\n\n\n\$mam_data\n" . Dumper($mam_data) . "\n";
#    close DBG;
    
    my $sl_graph = GD::Graph::lines->new($graph_width, $graph_height);    
    $sl_graph->set( %graph_options, 
        y_number_format => "%.0f",
        x_min_value => $bins[0],
        x_max_value => $bins[-1],
        y_max_value => ceil($sl_max/$num_y_ticks)*$num_y_ticks,
        l_margin => 26,
#        cumulate => 1,
#        accentclr => "black",
        dclrs => [ qw(seeders leechers total) ],   
    ) or do_log("mk_graphs: \$sl_graph->set returned error: " . $sl_graph->error);
    $sl_graph->set_legend("Seeders", "Total Peers (S+L)");
    $sl_graph->set_legend_font($fontname, 8);
    $sl_graph->set_x_axis_font($fontname, 8);
    $sl_graph->set_y_axis_font($fontname, 8);
    # $sl_graph->set_title_font($fontname, 12);


    my $cvh_graph = GD::Graph::lines->new($graph_width, $graph_height);    
    $cvh_graph->set( %graph_options, 
        y_number_format => "%.0f",
        x_min_value => $bins[0],
        x_max_value => $bins[-1],
        y_max_value => ceil($cvh_max/$num_y_ticks)*$num_y_ticks,
        l_margin => 18,
        dclrs => [ qw(completes views hits) ],   
    ) or do_log("mk_graphs: \$cvh_graph->set returned error: " . $cvh_graph->error);
    $cvh_graph->set_legend("Completes", "Views", "Hits");
    $cvh_graph->set_legend_font($fontname, 8);
    $cvh_graph->set_x_axis_font($fontname, 8);
    $cvh_graph->set_y_axis_font($fontname, 8);

    my $mam_graph = GD::Graph::lines->new($graph_width, $graph_height);    
    $mam_graph->set( %graph_options, 
        y_number_format => "%.0f%%",
        x_min_value => $bins[0],
        x_max_value => $bins[-1],
        y_max_value => 100,
        l_margin => 8,
        skip_undef => 1,
#        markers => [7, 7, 7],
#        marker_size => 2,
        dclrs => [ qw(min avg max) ],   
    ) or do_log("mk_graphs: \$mam_graph->set returned error: " . $mam_graph->error);
    $mam_graph->set_legend("Min", "Average", "Max");
    $mam_graph->set_legend_font($fontname, 8);
    $mam_graph->set_x_axis_font($fontname, 8);
    $mam_graph->set_y_axis_font($fontname, 8);

           
    open(IMG, ">", "$basename-sl.png") or do_log("mk_graphs: error writing $basename-sl.png: $!");
    binmode IMG;
    print IMG $sl_graph->plot($sl_data)->png or do_log("mk_graphs: error plotting \$sl_graph: " . $sl_graph->error);
    close IMG;

    open(IMG, ">", "$basename-cvh.png") or do_log("mk_graphs: error writing $basename-cvh.png: $!");
    binmode IMG;
    print IMG $cvh_graph->plot($cvh_data)->png or do_log("mk_graphs: error plotting \$cvh_graph: " . $cvh_graph->error);
    close IMG;

    open(IMG, ">", "$basename-mam.png") or do_log("mk_graphs: error writing $basename-mam.png: $!");
    binmode IMG;
    print IMG $mam_graph->plot($mam_data)->png or do_log("mk_graphs: error plotting \$mam_graph: " . $mam_graph->error);
    close IMG;

    my $time_end = [gettimeofday];

    do_log(sprintf("mk_graphs: %.3fs fetch / %.3fs gen / %ds %s / %d recs / id %d (%s)", 
                tv_interval($time_start, $time_fetch),
                tv_interval($time_fetch, $time_end),
                abs($secs_until),
                ($secs_until < 0 ? "behind" : "ahead"),
                $recno,
                $torrent_id, 
                $torrent_name,
                ));
    return $this_lsd;
}    


sub avg {
    # passed an array ref this returns its average
    my $aref = shift;
    return undef if(!defined($aref) || scalar(@$aref) < 1);

    my $tot = 0;
    
    $tot += $_ foreach @$aref;
    
    return scalar(@$aref) ? ($tot / scalar(@$aref)) : undef;
}

sub max {
    # passed an array ref this returns the maximum value
    my $aref = shift;

    return undef if(!defined($aref));
    
    my $max = undef;

    foreach( @$aref ) {
        $max = $_ if(!defined($max) || $_ > $max)
    }
    
#    print STDERR "max_debug: passed: \n" . Dumper($aref) . "returning: " . $max . "\n";
    return $max;
}
